#ifndef _TEXTUREDPLANE
#define _TEXTUREDPLANE

#include <CAVSharedObject.h>
#include <FLUtils.h>
#include "cave_ogl.h"

/** This class knows how to draw itself with OpenGL, and can use IRIX
    specific extensions. */

class TexturedPlane : public CAVSharedObject
{
public:
  /** @name Methods */
  //@{

  /// Constructor builds the TexturedPlane with minimal information

  TexturedPlane();

  /// Constructor builds the TexturedPlane with the specified information

  TexturedPlane(int width, int height, int depth);

  /// Destructor 

  ~TexturedPlane();

  /// Set the texture parameters, including the current data

  void setTexture(int width, int height, int depth, unsigned char *data, 
		  GLenum format = GL_RGB);

  /// Set the format of the Data

  void setFormat(int format);

  /// Graphics initialization 

  void appInit();
    
  /// Processing once per frame

  void frame();
    
  /// Drawing once per eye

  void draw();
    
  /// Get the texture width

  int GetWidth(void) {
    return textureWidth_;
  }

  /// Get the texture Height

  int GetHeight(void) {
    return textureHeight_;
  }

  /// Get the next power of two greater than the width

  int GetScaledWidth(void) {
    return nextpowtwowidth_;
  }

  /// Get the next power of two greater than the height

  int GetScaledHeight(void) {
    return nextpowtwoheight_;
  }

  /// We can spit out the pane data into and EPS file

  int writePaneEPS(char *filename);

protected:

  /// Get what wall we are supposed to draw on

  inline int getWallIndex(void)
  {
    for(int i = 0; i < CAVE_MAX_WALLS; i++)
      {
	if(CAVEWall == CAVEConfig->DisplayWall[i])
	  return i;
      }
    return -1;
  }

  /// Build the texture for the wall

  void buildWall(int);
    
  /// update the texture

  void rebuildWall(int);

  //@}

  /** @name Data */
  //@{

  /// Lock the data for display

  CAVELOCK lock_;

  /// Is this the first time we have built the wall

  int firstBuild_[CAVE_MAX_WALLS];

  /// 

  int dlIndex_[CAVE_MAX_WALLS];

  /// The current texture serial number

  int textureSerial_;

  /// The last serial number used on this wall

  int wallSerial_[CAVE_MAX_WALLS];

  /// The actual texture width

  int textureWidth_;

  /// The actual texture height

  int textureHeight_;

  /// The actual texture depth

  int textureDepth_;

  /// The next power of two greater than the width

  int nextpowtwowidth_;
  
  /// The next power of two greater than the height

  int nextpowtwoheight_;
  
  /// Someplace to store a frame temporarily

  unsigned char *tmpBuffer_;
  
  /// The actual texture data

  unsigned char *textureData_;

  /// The data format

  GLenum format_;
  //@}
};

#endif
