#ifndef _NETCAVECAM
#define _NETCAVECAM

/*
 *
 * This is part of the ANL/MCS/FL CAVEav distribution.
 * written by Ivan R. Judson
 *
 */


#include <GL/gl.h>
#include <GL/glu.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>

#include "CAVcam.h"

/** This is the derived class netCAVEcam. netCAVcam starts a vic
  process upon initialization which uses a shared memory object to
  retrieve the rendered frames from the netCAVcam.

  In order for this camera to find the vic binary it is necessary for
  the user to set the environment variable CAV_VIC to the full
  pathname (including trailing '/vic') to the vic binary that is
  compiled with shared memory support.  
*/

class netCAVcam : public CAVcam {

public:
  /**@name Methods */
  //@{
    
  /// Temporary constructor that takes optional arguments for the hostname and port of the video destination and optional arguments for the frame rate, bandwidth and time to live args.
	  
  netCAVcam(char *hostname = NULL, int port = 0,
	    int fr = 8, int bw = 128, int ttl = 4,
	    float ex = 0.0, float ey = 5.0, float ez = 0.0, 
	    float cx = 0.0, float cy = 5.0, float cz = -5.0,
	    float ux = 0.0, float uy = 1.0, float uz = 0.0);

    enum VideoFormat {
	JPEG,
	H261
    };
    void setVideoFormat(VideoFormat vf)
    {
	videoFormat_ = vf;

	if (videoFormat_ == H261)
	    setReflectImage(1);
    }

  /// This destructor kills the vic process and the shared memory segment

  ~netCAVcam();

  /// This initialization starts the vic process.

  void Init(void);

  /// This is the function that processes the pixels and hands them off to the next step.

  void PutPix(void);

    void setJpegQ(int q) { jpegQ_ = q; }
    void setH261Q(int q) { h261Q_ = q; }

  //@}
private:
  /**@name Data */
  //@{

  /// Shared Memory Key

  key_t _shmkey;

  /// Shared Memory Object

  FLSharedMemory *_shmem;

  /// Destination Hostname

  char *_host;

  /// Destination Port

  int _port;

  /// Vic process id

  int _pid;
  
  /// Vic frame rate

  int _frameRate;

  /// Vic bandwidth
  
  int _bandwidth;

  /// Vic Time To Live
  
  int _ttl;

  /// Camera configuration filename

  char *_fn;

   VideoFormat videoFormat_;

    int jpegQ_;
    int h261Q_;

  //@}
};

#endif

