/*
 *
 * This is part of the ANL/MCS/FL CAVEav distribution.
 * written by Ivan R. Judson
 *
 */

#include <netCAVcam.h>

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#define CONFDATA "option add Vic.bandwidth %d startupFile \
\noption add Vic.mtu 4096 startupFile \
\noption add Vic.framerate %d startupFile \
\noption add Vic.defaultDevice shmem \
\noption add Vic.transmitOnStartup True\
\noption add Vic.muteNewSources True\
\n\nproc user_hook {} { \
\n\tglobal qscale_val \
\n\tglobal inputDeviceList \
\n\tset qscale_val(jpeg) %d \
\n\tset qscale_val(h261) %d \
\n\tforeach v $inputDeviceList { \
\n\t\tif { [$v nickname] == \"shmem\" } { \
\n\t\t\t$v id %d \n\t\t} \n\t}\n} "


netCAVcam::netCAVcam(char *hostname, int port,
                     int fr, int bw, int ttl,
                     float ex, float ey, float ez, 
                     float cx, float cy, float cz, 
                     float ux, float uy, float uz) : CAVcam(ex, ey, ez,
                                                            cx, cy, cz,
                                                            ux, uy, uz)
{
  time_t tloc;
  time(&tloc);
  srand48((long)tloc);
  
  while(_shmkey < 1)
    _shmkey = lrand48() % 32768;

#ifdef DEBUG
  fprintf(stderr, "Key: %d", _shmkey);
#endif

  setJpegQ(79);
  setH261Q(68);

  _host = hostname;
  _port = port;

  _fn = tempnam("/tmp/","CAM");

  _frameRate = fr;
  _bandwidth = bw;
  _ttl = ttl;

  _shmem = new FLSharedMemory(_shmkey, sizeof(FLImagePool), 0666);

  _format = GL_ABGR_EXT;

  _imagepool = (FLImagePool *)_shmem->GetAddress();

  _imagepool->_nImages = 10;
  _imagepool->_totalPoolSize = sizeof(FLImagePool);
  _imagepool->_nextImage = _imagepool->_currentSerial = -1;
  _imagepool->Resize(320, 240, 4);

#ifdef VIEWFINDER
  viewFinder_ = new TexturedPlane(320, 240, 4);
  //  viewFinder_->Init();
#endif

}

netCAVcam::~netCAVcam()
{
  delete _shmem;

  unlink(_fn);
  kill(_pid, SIGKILL);
}

void netCAVcam::Init(void)
{
  FILE *fp;
  char cmdpath[256];
  char cmdargs[2][128];
  char vicconf[1024];
  char *envpath;
  
#ifdef DEBUG
  fprintf(stderr, "Entering netCAVcam Init function.\n");
#endif

#if defined(GLXPIXMAP) || defined(PBUFFER)
  CAVcam::setupOffscreen();
#endif

  sprintf(cmdpath,"%s",  VICPATH);

  if((envpath = getenv("CAV_VIC")) != NULL)
    strcpy(cmdpath, envpath);

#ifdef DEBUG
  fprintf(stderr, "Opening vic configuration file...");
#endif

  if((fp = fopen(_fn, "w")) != NULL) 
    {
      sprintf(vicconf, CONFDATA, _bandwidth, _frameRate,
	      jpegQ_, h261Q_,
	      _shmem->GetID());
      fprintf(fp, "%s\n", vicconf);
      fclose(fp);
    } else {
      perror("Can't open file");
      CAVEExit();
    }

#ifdef DEBUG
  fprintf(stderr, "done.\nBuilding vic command line...");
#endif

  char *formatString;
  switch (videoFormat_)
  {
  case JPEG:
      formatString = "jpeg";
      break;

  case H261:
      formatString = "h261";
      break;
  }

  sprintf(cmdargs[0], "-u %s", _fn);
  sprintf(cmdargs[1], "%s/%d/%s/%d", _host, _port, formatString, _ttl);

#ifdef DEBUG
  fprintf(stderr, "done.\nForking vic process...");
#endif
  _pid = fork();

  if(!_pid)
    {
#if defined(DEBUG) || 1
      fprintf(stderr, "done.\nExecing vic...");
      fprintf(stderr, "%s %s %s\n", cmdpath, cmdargs[0], cmdargs[1]);

      FILE *fp = fopen("/tmp/vic.sh", "w");
      if (fp == 0)
	  perror("Cannot open /tmp/vic.sh");
      else
      {
	  fprintf(fp, "#!/bin/sh -x\n");
	  fprintf(fp, "%s %s %s\n", cmdpath, cmdargs[0], cmdargs[1]);
	  fchmod(fileno(fp), 0755);
	  fclose(fp);
      }

#endif

      if (getenv("NO_VIC") == 0)
      {
	  if(execl(cmdpath, cmdpath, cmdargs[0], cmdargs[1], NULL) == -1) 
	  {
	      perror("exec failed");
	      CAVEExit();
	  }
      }
#ifdef DEBUG
      fprintf(stderr, "done.\n");
#endif
      exit(0);
    }

#ifdef DEBUG
  fprintf(stderr, "Successfully exiting netCAVcam Init function.\n");
#endif
}

void netCAVcam::PutPix()
{
  // turn the pixels around
//  FlipFrame(_imagepool->_nextImage);
}

