/*
 * Copyright (c) 1993-1995 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: /home/FuturesLab/Software/CVSMaster/CAVEav/vic-mods/grabber-shm.cc,v 1.8 1998/08/08 01:21:13 olson Exp $ (LBL)";
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>
#ifndef WIN32
#include <sys/file.h>
#endif
#include <sys/stat.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include "inet.h"
#include "Tcl.h"
#include "rtp.h"
#include "grabber.h"
#include "crdef.h"
#include "device-input.h"
#include "transmitter.h"
#include "module.h"
#include "rgb-converter.h"
#include "FLImagePool.h"
#include "FLSharedMemory.h"

#ifdef __sgi
#include "JpegEncoder.h"
#endif

#define WID  320
#define HEI  240
#define DEP  4
#define FSIZE (WID*HEI*DEP)

//#define GSDEBUG

class ShmemDevice : public InputDevice {
public:
  ShmemDevice(const char* s);
  virtual int command(int argc, const char * const * argv);
  void *Address(void) { return _cshm->GetAddress(); }
  int isAttached(void) { return attached; }
private:
  void attach_shmem(int port);
  int attached;
  FLSharedMemory *_cshm;
};

class ShmemGrabber : public Grabber 
{
public:
  ShmemGrabber(ShmemDevice &device);
  ~ShmemGrabber();
  virtual int command(int argc, const char* const * argv);

protected:
  void start();
  void stop();
    virtual int grab();
  ShmemDevice &device_;
  FLImagePool *frames_;
  RGB_Converter *converter_;
  int _lastserial;
  int decimate_;
};

#ifdef __sgi
class ShmemJpegGrabber : public ShmemGrabber {
public:
    ShmemJpegGrabber(ShmemDevice &device);
    ~ShmemJpegGrabber();
    int command(int argc, const char* const * argv);
    int grab();
    void setQ(int q);

private:
    int q_;
    JpegEncoder *jpegEncoder_;
};
#endif

class Shmem422Grabber : public ShmemGrabber {
public:
  Shmem422Grabber(ShmemDevice &device);
  ~Shmem422Grabber();
  int command(int argc, const char* const * argv);
};

class Shmem411Grabber : public ShmemGrabber {
public:
  Shmem411Grabber(ShmemDevice &device);
  ~Shmem411Grabber();
  int command(int argc, const char* const * argv);
};

class ShmemCIFGrabber : public ShmemGrabber {
public:
  ShmemCIFGrabber(ShmemDevice &device);
  ~ShmemCIFGrabber();
  int command(int argc, const char* const * argv);
};

static ShmemDevice Shmem_device("shmem");

ShmemDevice::ShmemDevice(const char* s) : InputDevice(s)
{
#ifdef __sgi
    attributes_ = "\
format { 422 411 CIF jpeg } \
size { small large cif } \
";
#else
    attributes_ = "\
format { 422 411 CIF } \
size { small large cif } \
";
#endif    
    attached = 0;
}

int ShmemDevice::command(int argc, const char*const* argv)
{
#ifdef GSDEBUG
	  fprintf(stderr,"ShmemDevice::command argc=%d ", argc);
	  int i;
	  for (i = 0; i < argc; i++)
	    fprintf(stderr,"\"%s\" ", argv[i]);
	  fprintf(stderr,"\n");
#endif 
    if (argc == 3)
    {
	if (strcmp(argv[1], "open") == 0)
	{
	    const char* fmt = argv[2];
	    TclObject* o = 0;

	    if (strcmp(fmt, "422") == 0)
	      o = new Shmem422Grabber(*this);
  	    else if (strcmp(fmt, "411") == 0)
	      o = new Shmem411Grabber(*this);
	    else if (strcmp(fmt, "cif") == 0)
	      o = new ShmemCIFGrabber(*this);
	    else if (strcmp(fmt, "jpeg") == 0)
	      o = new ShmemJpegGrabber(*this);
	    
	    if (o != 0)
		Tcl::instance().result(o->name());

	    return (TCL_OK);
	} else if (strcmp(argv[1], "id") == 0)
	  {
	    attach_shmem(atoi(argv[2]));
	    return (TCL_OK);
	  }
    }
    return (InputDevice::command(argc, argv));
}

void ShmemDevice::attach_shmem(int id)
{
  _cshm = new FLSharedMemory(id);
  attached = 1;
}

// Grabbers 
ShmemGrabber::ShmemGrabber(ShmemDevice &device) : device_(device)
{
  decimate_ = 1;
  _lastserial = -1;
  frame_ = new u_int8_t[FSIZE];
}

ShmemGrabber::~ShmemGrabber()
{
#ifdef GSDEBUG
    fprintf(stderr,"Destroy ShmemGrabber\n");
#endif
}

int ShmemGrabber::command(int argc, const char * const * argv)
{
    Tcl& tcl = Tcl::instance();
    
#ifdef GSDEBUG
    fprintf(stderr,"ShmemGrabber::command argc=%d ", argc);
    int i;
    for (i = 0; i < argc; i++)
	fprintf(stderr,"\"%s\" ", argv[i]);
    fprintf(stderr,"\n");
#endif /* GSDEBUG */

    if (argc == 3)
    {
      if (strcmp(argv[1], "q") == 0)
	{
	  return TCL_OK;
	} 
    }

    return (Grabber::command(argc, argv));
}

void ShmemGrabber::start()
{
    frameclock_ = gettimeofday();

    timeout();
}

void ShmemGrabber::stop()
{
    cancel();
}

int ShmemGrabber::grab()
{
  int thisSerial;
  FLImageDescriptor *image;

  if(device_.isAttached() && frames_ != NULL)
    {
      thisSerial = frames_->_currentSerial;

      if(thisSerial > _lastserial)
	{
	  FLImageDescriptor *image;
      
	  image = &(frames_->_images[frames_->_nextImage]);

	  converter_->convert((u_int8_t *)image->GetData(), image->GetWidth(), 
			      image->GetHeight(), frame_);
	  suppress(frame_);
	  saveblks(frame_);
	  YuvFrame f(media_ts(), frame_, crvec_, outw_, outh_);
	  _lastserial = thisSerial;
	  
	  return(target_->consume(&f));
	} else {
	  return 0;
	}
    }
  return 0;
}

// 422 Grabber

Shmem422Grabber::Shmem422Grabber(ShmemDevice &device) : ShmemGrabber(device)
{
  device_ = device;
  frames_ = (FLImagePool *)device.Address();
  decimate_ = 1;
  converter_ = RGB_Converter_422::instance();
  set_size_422(WID, HEI);
  allocref();
}

int Shmem422Grabber::command(int argc, const char * const * argv)
{
    Tcl& tcl = Tcl::instance();
    
#ifdef GSDEBUG
    fprintf(stderr,"Shmem422Grabber::command argc=%d ", argc);
    int i;
    for (i = 0; i < argc; i++)
	fprintf(stderr,"\"%s\" ", argv[i]);
    fprintf(stderr,"\n");
#endif /* GSDEBUG */

    if (argc == 3)
    {
      if(strcmp(argv[1], "decimate") == 0)
	{
	  decimate_ = atoi(argv[2]);
	  if(device_.isAttached() && frames_ != NULL)
	    {
	      frames_->_clientActive = 0;
	      frames_->Resize(WID/decimate_, HEI/decimate_, DEP);
	      frames_->_clientActive = 1;
	    } else {
	      frames_ = (FLImagePool *)device_.Address();
	      frames_->Resize(WID/decimate_, HEI/decimate_, DEP);
	      frames_->_clientActive = 1;
	    }	      
	  set_size_422(WID/decimate_, HEI/decimate_);
	  allocref();
	}
      else if (strcmp(argv[1], "q") == 0)
	{
	  return TCL_OK;
	}
    }

    return (ShmemGrabber::command(argc, argv));
}

Shmem422Grabber::~Shmem422Grabber()
{
  delete converter_;

#ifdef GSDEBUG
    fprintf(stderr,"Destroy Shmem422Grabber\n");
#endif
}

// 411 Grabber

Shmem411Grabber::Shmem411Grabber(ShmemDevice &device) : ShmemGrabber(device)
{
  device_ = device;
  frames_ = (FLImagePool *)device.Address();
  decimate_ = 1;
  converter_ = RGB_Converter_411::instance();
  set_size_411(WID, HEI);
  allocref();
}

int Shmem411Grabber::command(int argc, const char * const * argv)
{
    Tcl& tcl = Tcl::instance();
    
#ifdef GSDEBUG
    fprintf(stderr,"Shmem411Grabber::command argc=%d ", argc);
    int i;
    for (i = 0; i < argc; i++)
	fprintf(stderr,"\"%s\" ", argv[i]);
    fprintf(stderr,"\n");
#endif /* GSDEBUG */

    if (argc == 3)
    {
      if(strcmp(argv[1], "decimate") == 0)
	{
	  decimate_ = atoi(argv[2]);
	  if(device_.isAttached() && frames_ != NULL)
	    {
	      frames_->_clientActive = 0;
	      frames_->Resize(WID/decimate_, HEI/decimate_, DEP);
	      frames_->_clientActive = 1;
	    } else {
	      frames_ = (FLImagePool *)device_.Address();
	      frames_->Resize(WID/decimate_, HEI/decimate_, DEP);
	      frames_->_clientActive = 1;
	    }
	  set_size_411(WID/decimate_, HEI/decimate_);
	  allocref();
	  return TCL_OK;	
	}
      else if (strcmp(argv[1], "q") == 0)
	{
	    return TCL_OK;
	}
    }

    return (ShmemGrabber::command(argc, argv));
}

Shmem411Grabber::~Shmem411Grabber()
{
//  delete converter_;

#ifdef GSDEBUG
    fprintf(stderr,"Destroy Shmem411Grabber\n");
#endif
}

// CIF Grabber
ShmemCIFGrabber::ShmemCIFGrabber(ShmemDevice &device) : ShmemGrabber(device)
{
  device_ = device;
  frames_ = (FLImagePool *)device.Address();
  decimate_ = 1;
  converter_ = RGB_Converter_411::instance();
  set_size_411(WID, HEI);
  allocref();
}

int ShmemCIFGrabber::command(int argc, const char * const * argv)
{
    Tcl& tcl = Tcl::instance();
    
#ifdef GSDEBUG
    fprintf(stderr,"ShmemCIFGrabber::command argc=%d ", argc);
    int i;
    for (i = 0; i < argc; i++)
	fprintf(stderr,"\"%s\" ", argv[i]);
    fprintf(stderr,"\n");
#endif /* GSDEBUG */

    if (argc == 3)
    {
      if(strcmp(argv[1], "decimate") == 0)
	{
	  decimate_ = atoi(argv[2]);
	  if(device_.isAttached() && frames_ != NULL)
	    {
	      frames_->_clientActive = 0;
	      frames_->Resize((2 * 352)/decimate_, (2 * 288)/decimate_, DEP);
	      frames_->_clientActive = 1;
	    } else {
	      frames_ = (FLImagePool *)device_.Address();
	      frames_->Resize((2 * 352)/decimate_, (2 * 288)/decimate_, DEP);
	      frames_->_clientActive = 1;
	    }
	  set_size_411((2 * 352)/decimate_, (2 * 288)/decimate_);
	  allocref();
	  return TCL_OK;
	}
	else if (strcmp(argv[1], "q") == 0)
	{
	    return TCL_OK;
	}
    }
    return (ShmemGrabber::command(argc, argv));
}

ShmemCIFGrabber::~ShmemCIFGrabber()
{
//  delete converter_;

#ifdef GSDEBUG
    fprintf(stderr,"Destroy ShmemCIFGrabber\n");
#endif
}

#ifdef __sgi
// Jpeg Grabber
ShmemJpegGrabber::ShmemJpegGrabber(ShmemDevice &device) : ShmemGrabber(device)
{
    device_ = device;
    frames_ = (FLImagePool *)device.Address();
    decimate_ = 1;
    int w, h;
    frames_->getFrameSize(w, h);
    q_ = 75;
    jpegEncoder_ = new JpegEncoder(w, h, DM_IMAGE_PACKING_XBGR, q_);
}

ShmemJpegGrabber::~ShmemJpegGrabber()
{
    delete jpegEncoder_;
#ifdef GSDEBUG
    fprintf(stderr,"Destroy ShmemJpegGrabber\n");
#endif
}

int ShmemJpegGrabber::grab()
{
    int thisSerial;
    FLImageDescriptor *image;

    if(device_.isAttached() && frames_ != NULL)
    {
	thisSerial = frames_->_currentSerial;

	if(thisSerial <= _lastserial)
	{
	    return 0;
	}
    }
    else
	return 0;
      
    image = &(frames_->_images[frames_->_nextImage]);

    // image data in image->GetData()
    // size in image->GetWidth(), image->GetHeight()
    //
    // frame_ is the local buffer for the frame
    //

    void *jpegFrame;
    int size;
    DMbuffer buffer;
    jpegEncoder_->convertFrame((void *) image->GetData(), &buffer);
    jpegFrame = dmBufferMapData(buffer);
    size = dmBufferGetSize(buffer);
//    memcpy(frame_, jpegFrame, size);

    static int wrote = 0;

    if (!wrote)
    {
	printf("Writing frame\n");
	FILE *fp = fopen("out.jpg", "w");
	fwrite(jpegFrame, size, 1, fp);
	fclose(fp);
	wrote = 1;
    }

    /* strip off jfif header */
    int sawff = 0;
    u_char* ep;
    u_char *p = (u_char *) jpegFrame;
//    u_char *p = (u_char *) frame_; 
    for (ep = p + size; p < ep; )
    {
	int c = *p++;
	if (c == 0xff) {
	    sawff = 1;
	    continue;
	}
	if (!sawff)
	    continue;
	if  (c == 0xda) {
	    /* found start-of-scan marker */
	    if (p + 2 <= ep) {
		/* skip over SOS */
		u_int t = (p[0] << 8) | p[1];
		p += t;
		break;
	    }
	}
	sawff = 0;
    }

    if (p >= ep)
	return (0);

    JpegFrame f(media_ts(), p, ep - p, q_, 0,
		image->GetWidth(), image->GetHeight());

    _lastserial = thisSerial;
    int rc = (target_->consume(&f));
    dmBufferFree(buffer);

    return rc;
}

void ShmemJpegGrabber::setQ(int q)
{
    q_ = q;
    jpegEncoder_->setQuality(q_);
}

int ShmemJpegGrabber::command(int argc, const char * const * argv)
{
    Tcl& tcl = Tcl::instance();
    
#ifdef GSDEBUG
    fprintf(stderr,"ShmemJpegGrabber::command argc=%d ", argc);
    int i;
    for (i = 0; i < argc; i++)
	fprintf(stderr,"\"%s\" ", argv[i]);
    fprintf(stderr,"\n");
#endif /* GSDEBUG */

    if (argc == 3)
    {
	if (strcmp(argv[1], "q") == 0)
	{
	    int q = atoi(argv[2]);	    
	    setQ(q);
	    return TCL_OK;
	}
    }
    return (ShmemGrabber::command(argc, argv));
}

#endif /* __sgi */
